@@ -12,9 +12,10 @@ See svg.py and aa.py for output modules, that can render the parsed structure.
 This is open source software under the BSD license. See LICENSE.txt for more
 details.
 """
-
+import codecs
 from error import UnsupportedFormatError
 from shapes import *
+from unicodedata import east_asian_width
 import sys
 
 NOMINAL_SIZE = 2
@@ -35,6 +36,8 @@ DEFAULT_OPTIONS = dict(
     debug        = False,
     textual      = False,
     proportional = False,
+    encoding     = 'utf-8',
+    widechars     = 'F,W',
 )
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -56,7 +59,7 @@ class AsciiArtImage:
 
     QUOTATION_CHARACTERS = list('"\'`')
 
-    def __init__(self, text, aspect_ratio=1, textual=False):
+    def __init__(self, text, aspect_ratio=1, textual=False, widechars='F,W'):
         """Take a ASCII art figure and store it, prepare for ``recognize``"""
         self.aspect_ratio = float(aspect_ratio)
         self.textual = textual
@@ -65,9 +68,20 @@ class AsciiArtImage:
         self.image = []
         max_x = 0
         y = 0
+        # define character widths map
+        charwidths = {}
+        for key in ['F', 'H', 'W', 'Na', 'A', 'N']:
+            if key in widechars.split(','):
+                charwidths[key] = 2
+            else:
+                charwidths[key] = 1
         for line in text.splitlines():
-            max_x = max(max_x, len(line))
-            self.image.append(line)
+            # extend length by 1 for each wide glyph
+            line_len = sum(charwidths[east_asian_width(c)] for c in line)
+            max_x = max(max_x, line_len)
+            # pad a space for each wide glyph
+            padded_line = ''.join(c+' '*(charwidths[east_asian_width(c)]-1) for c in line)
+            self.image.append(padded_line)
             y += 1
         self.width = max_x
         self.height = y
@@ -767,28 +781,38 @@ class AsciiArtImage:
         result = []
         if self.get(x, y) == '/':
             # rounded rectangles
-            # XXX return Arc shapes, not Lines
             if (self.get(x + 1, y) == '-' and self.get(x, y + 1) == '|'):
-                result.append(Line(
-                    Point(self.hcenter(x), self.bottom(y)),
-                    Point(self.right(x), self.vcenter(y))
+                # upper left corner
+                result.append(Arc(
+                    Point(self.hcenter(x), self.bottom(y)), 90,
+                    Point(self.right(x), self.vcenter(y)),  180
                 ))
             if self.get(x - 1, y) == '-' and self.get(x, y - 1) == '|':
-                result.append(Line(
-                    Point(self.hcenter(x), self.top(y)),
-                    Point(self.left(x), self.vcenter(y))
+                # lower right corner
+                result.append(Arc(
+                    Point(self.hcenter(x), self.top(y)),  -90,
+                    Point(self.left(x), self.vcenter(y)), 0
                 ))
             if not result:
                 # if used as diagonal line
                 p1 = p2 = None
+                a1 = a2 = 0
+                arc = c1 = c2 = False
                 if self.get(x + 1, y - 1) == '|':
                     p1 = Point(self.hcenter(x + 1), self.top(y))
+                    a1 = -90
+                    arc = c1 = True
                 elif self.get(x + 1, y - 1) == '+':
                     p1 = Point(self.hcenter(x + 1), self.vcenter(y - 1))
+                    a1 = -135
                 elif self.get(x + 1, y - 1) == '-':
                     p1 = Point(self.right(x), self.vcenter(y - 1))
+                    a1 = 180
+                    arc = c1 = True
                 elif self.get(x + 1, y - 1) == '/':
                     p1 = Point(self.right(x), self.top(y))
+                    a1 = -135
+                    c1 = True
                 elif self.get(x + 1, y) == '|':
                     p1 = Point(self.hcenter(x + 1), self.top(y))
                 elif self.get(x, y - 1) == '-':
@@ -796,12 +820,19 @@ class AsciiArtImage:
 
                 if self.get(x - 1, y + 1) == '|':
                     p2 = Point(self.hcenter(x - 1), self.top(y + 1))
+                    a2 = 90
+                    arc = c2 = True
                 elif self.get(x - 1, y + 1) == '+':
                     p2 = Point(self.hcenter(x - 1), self.vcenter(y + 1))
+                    a2 = 45
                 elif self.get(x - 1, y + 1) == '-':
                     p2 = Point(self.left(x), self.vcenter(y + 1))
+                    a2 = 0
+                    arc = c2 = True
                 elif self.get(x - 1, y + 1) == '/':
                     p2 = Point(self.left(x), self.bottom(y))
+                    a2 = 45
+                    c2 = True
                 elif self.get(x - 1, y) == '|':
                     p2 = Point(self.hcenter(x - 1), self.bottom(y))
                 elif self.get(x, y + 1) == '-':
@@ -812,30 +843,44 @@ class AsciiArtImage:
                         p1 = Point(self.right(x), self.top(y))
                     if not p2:
                         p2 = Point(self.left(x), self.bottom(y))
-                    result.append(Line(p1, p2))
+                    if arc:
+                        result.append(Arc(p1, a1, p2, a2, c1, c2))
+                    else:
+                        result.append(Line(p1, p2))
         else: # '\'
             # rounded rectangles
             if self.get(x-1, y) == '-' and self.get(x, y + 1) == '|':
-                result.append(Line(
-                    Point(self.hcenter(x), self.bottom(y)),
-                    Point(self.left(x), self.vcenter(y))
+                # upper right corner
+                result.append(Arc(
+                    Point(self.hcenter(x), self.bottom(y)), 90,
+                    Point(self.left(x), self.vcenter(y)),   0
                 ))
             if self.get(x+1, y) == '-' and self.get(x, y - 1) == '|':
-                result.append(Line(
-                    Point(self.hcenter(x), self.top(y)),
-                    Point(self.right(x), self.vcenter(y))
+                # lower left corner
+                result.append(Arc(
+                    Point(self.hcenter(x), self.top(y)),   -90,
+                    Point(self.right(x), self.vcenter(y)), 180
                 ))
             if not result:
                 # if used as diagonal line
                 p1 = p2 = None
+                a1 = a2 = 0
+                arc = c1 = c2 = False
                 if self.get(x - 1, y - 1) == '|':
                     p1 = Point(self.hcenter(x-1), self.top(y))
+                    a1 = -90
+                    arc = c1 = True
                 elif self.get(x - 1, y - 1) == '+':
                     p1 = Point(self.hcenter(x-1), self.vcenter(y - 1))
+                    a1 = -45
                 elif self.get(x - 1, y - 1) == '-':
                     p1 = Point(self.left(x), self.vcenter(y-1))
+                    a1 = 0
+                    arc = c1 = True
                 elif self.get(x - 1, y - 1) == '\\':
                     p1 = Point(self.left(x), self.top(y))
+                    a1 = -45
+                    c1 = True
                 elif self.get(x - 1, y) == '|':
                     p1 = Point(self.hcenter(x-1), self.top(y))
                 elif self.get(x, y - 1) == '-':
@@ -843,12 +888,19 @@ class AsciiArtImage:
 
                 if self.get(x + 1, y + 1) == '|':
                     p2 = Point(self.hcenter(x+1), self.top(y + 1))
+                    a2 = 90
+                    arc = c2 = True
                 elif self.get(x + 1, y + 1) == '+':
                     p2 = Point(self.hcenter(x+1), self.vcenter(y + 1))
+                    a2 = 135
                 elif self.get(x + 1, y + 1) == '-':
                     p2 = Point(self.right(x), self.vcenter(y + 1))
+                    a2 = 180
+                    arc = c2 = True
                 elif self.get(x + 1, y + 1) == '\\':
                     p2 = Point(self.right(x), self.bottom(y))
+                    a2 = 135
+                    c2 = True
                 elif self.get(x + 1, y) == '|':
                     p2 = Point(self.hcenter(x+1), self.bottom(y))
                 elif self.get(x, y + 1) == '-':
@@ -859,7 +911,10 @@ class AsciiArtImage:
                         p1 = Point(self.left(x), self.top(y))
                     if not p2:
                         p2 = Point(self.right(x), self.bottom(y))
-                    result.append(Line(p1, p2))
+                    if arc:
+                        result.append(Arc(p1, a1, p2, a2, c1, c2))
+                    else:
+                        result.append(Line(p1, p2))
         if result:
             self.tag([(x, y)], CLASS_JOIN)
         return group(result)
@@ -903,7 +958,7 @@ def process(input, visitor_class, options=None):
     if options['debug']:
         sys.stderr.write('%r\n' % (input,))
 
-    aaimg = AsciiArtImage(input, options['aspect'], options['textual'])
+    aaimg = AsciiArtImage(input, options['aspect'], options['textual'], options['widechars'])
 
     if options['debug']:
         sys.stderr.write('%s\n' % (aaimg,))
@@ -988,9 +1043,9 @@ def main():
     parser = optparse.OptionParser(
         usage = "%prog [options] [file]",
         version = """\
-%prog 0.3
+%prog 0.5
 
-Copyright (C) 2006-2009 aafigure-team
+Copyright (C) 2006-2010 aafigure-team
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License.
@@ -1009,6 +1064,20 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
     description = "ASCII art to image (SVG, PNG, JPEG, PDF and more) converter."
     )
 
+    parser.add_option("-e", "--encoding",
+        dest = "encoding",
+        action = "store",
+        help = "character encoding of input text",
+        default = DEFAULT_OPTIONS['encoding'],
+    )
+
+    parser.add_option("-w", "--wide-chars",
+        dest = "widechars",
+        action = "store",
+        help = "unicode properties to be treated as wide glyph (e.g. 'F,W,A')",
+        default = DEFAULT_OPTIONS['widechars'],
+    )
+
     parser.add_option("-o", "--output",
         dest = "output",
         metavar = "FILE",
@@ -1052,7 +1121,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
     )
 
     parser.add_option("-l", "--linewidth",
-        dest = "linewidth",
+        dest = "line_width",
         action = "store",
         type = 'float',
         help = "set width, svg only",
@@ -1105,9 +1174,10 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
             options.format = os.path.splitext(options.output)[1][1:]
 
     if args:
-        input = file(args[0])
+        _input = file(args[0])
     else:
-        input = sys.stdin
+        _input = sys.stdin
+    input = codecs.getreader(options.encoding)(_input)
 
     if options.output is None:
         output = sys.stdout
@@ -1116,8 +1186,8 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
     # explicit copying of parameters to the options dictionary
     options_dict = {}
-    for key in ('textual', 'proportional',
-                'linewidth', 'aspect', 'scale',
+    for key in ('widechars', 'textual', 'proportional',
+                'line_width', 'aspect', 'scale',
                 'format', 'debug'):
         options_dict[key] = getattr(options, key)
     # ensure all color parameters start with a '#'
